/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_SCHEDULER_H_
#define AIDGE_SCHEDULER_H_

#include <chrono>
#include <memory>
#include <set>
#include <string>
#include <vector>
#include <map>

#include "aidge/utils/Types.h"

#include "aidge/data/Tensor.hpp"
#include "aidge/scheduler/MemoryManager.hpp"

namespace Aidge {
class Node;
class GraphView;

class Scheduler {
protected:
    struct StaticSchedulingElement {
        StaticSchedulingElement(
            std::shared_ptr<Node> node_,
            size_t early_ = static_cast<size_t>(-1),
            size_t late_ = static_cast<size_t>(-1))
            : node(node_), early(early_), late(late_) {}

        std::shared_ptr<Node> node;
        size_t early;
        size_t late;
        std::vector<std::shared_ptr<StaticSchedulingElement>> earlierThan;
        std::vector<std::shared_ptr<StaticSchedulingElement>> laterThan;
    };

    struct SchedulingElement {
        SchedulingElement(
            std::shared_ptr<Node> node_,
            std::chrono::time_point<std::chrono::high_resolution_clock> start_,
            std::chrono::time_point<std::chrono::high_resolution_clock> end_)
            : node(node_), start(start_), end(end_) {}

        std::shared_ptr<Node> node;
        std::chrono::time_point<std::chrono::high_resolution_clock> start;
        std::chrono::time_point<std::chrono::high_resolution_clock> end;
    };

    struct PriorProducersConsumers {
        bool isPrior = false;
        std::set<std::shared_ptr<Aidge::Node>> requiredProducers;
        std::set<std::shared_ptr<Aidge::Node>> priorConsumers;
    };

public:
    Scheduler(std::shared_ptr<GraphView> graphView, std::shared_ptr<Node> upperNode = nullptr)
        : mGraphView(graphView),
          mUpperNode(upperNode)
    {
        // ctor
    };
    virtual ~Scheduler() = default;

    /**
     * Generate full static scheduling of the GraphView.
     * For each node, an earliest and latest possible execution logical step
     * is specified. Nodes that may be scheduled at the same logical step have
     * no data dependency and can be run in parallel.
    */
    void generateScheduling();

    /**
     * Reset all scheduling and associated nodes producer consumer.
    */
    void resetScheduling();

    /**
     * Generate the memory layout for the current static scheduling.
     * @param incProducers If true, include the producers in the memory layout.
     * @param wrapAroundBuffer If true, allow wrapping in memory planes.
    */
    MemoryManager generateMemory(bool incProducers = false, bool wrapAroundBuffer = false) const;

    /**
     * @brief Place the data tensors inside in the data input tensor of the graphView. In case of multiple data input tensors, they are mapped to producers in the order given by the graph.
     *
     * @param data data input tensors
     */
    void connectInputs(std::vector<std::shared_ptr<Aidge::Tensor>> data);

    /**
     * @brief Save in a Markdown file the static scheduling with early and late relative order for the nodes.
     * @param fileName Name of the generated file.
     */
    void saveStaticSchedulingDiagram(const std::string& fileName) const;

    /**
     * @brief Save in a Markdown file the order of layers execution.
     * @param fileName Name of the generated file.
     */
    void saveSchedulingDiagram(const std::string& fileName) const;

    /**
     * @brief Return a vector of Node ordered by the order they are called by the scheduler
     * @return std::vector<std::shared_ptr<Node>>
     */
    std::vector<std::shared_ptr<Node>> getStaticScheduling(size_t step = 0) const;
    inline std::shared_ptr<GraphView> getGraphView() const noexcept {
        return mGraphView;
    }

protected:
    /**
     * Generate an initial base scheduling for the GraphView.
     * The scheduling is entirely sequential and garanteed to be valid w.r.t.
     * each node producer-consumer model.
    */
    std::vector<std::shared_ptr<StaticSchedulingElement>> generateBaseScheduling() const;

    /**
     * Fill-in early and late scheduling step from initial base scheduling.
     * For each node, specifies the earliest and latest possible execution 
     * logical step.
    */
    void generateEarlyLateScheduling(std::vector<std::shared_ptr<StaticSchedulingElement>>& schedule) const;

    /**
     * @brief Set of layers receiving an input from currently processing layers
     *
     * @param producers Set of layers ready to run.
     * @return std::set<std::shared_ptr<Node>>
     */
    std::set<std::shared_ptr<Node>> getConsumers(const std::set<std::shared_ptr<Node>>& producers) const;
    Elts_t getNbAvailableData(const std::shared_ptr<Node>& node, const IOIndex_t inputIdx) const;
    PriorProducersConsumers getPriorProducersConsumers(const std::shared_ptr<Node>& node) const;

    /** @brief Shared ptr to the scheduled graph view */
    std::shared_ptr<GraphView> mGraphView;
    /** @brief Shared ptr to the upper node containing the graph view */
    std::weak_ptr<Node> mUpperNode;
    /** @brief List of SchedulingElement (i.e: Nodes with their computation time) */
    std::vector<SchedulingElement> mScheduling;
    /** @brief List of nodes ordered by their */
    std::vector<std::vector<std::shared_ptr<StaticSchedulingElement>>> mStaticSchedule;
    size_t mStaticScheduleStep = 0;
    mutable std::map<std::shared_ptr<Node>, PriorProducersConsumers> mPriorCache;
};
} // namespace Aidge

#endif /* AIDGE_SCHEDULER_H_ */
